/* MoarNES source - PPU.c

   This open-source software is (c)2010-2012 Mike Chambers, and is released
   under the terms of the GNU GPL v2 license.

   This is my implementation of the Picture Processing Unit (PPU) of the
   Nintendo Entertainment System. It is not perfect at this point, but the
   vast majority games will run correctly.

   This is arguably the most difficult aspect of a NES emulator to make
   perfect!
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "stdint.h"
#include "PPU.h"
#include "rom.h"

struct PPU_s *PPU;
struct OAM_s *OAM;
uint8_t *VRAM;
uint8_t outputNES[240][256];

extern struct cart_s cartridge;
extern uint8_t outputNES[240][256];

extern struct map9_s *map9;

extern void nmi6502();
extern void CHRswap(struct cart_s *cart, uint16_t banknum, uint16_t newbank, uint16_t banksize);
extern void fatalerr(uint8_t *errmsg);

extern void map4irqdecrement();

uint16_t lastaddr = 0;
uint8_t readPPU(uint16_t addr) {
	uint8_t tempbyte, oldbyte;
	addr &= 0x3FFF;

	//if (!(lastaddr & 0x1000) && (addr & 0x1000)) map4irqdecrement();

	if (addr>=0x3F00) addr = 0x3F00 | (addr & 0x1F);
	if (addr==0x3F10) addr = 0x3F00;
	if (addr==0x3F14) addr = 0x3F04;
	if (addr==0x3F18) addr = 0x3F08;
	if (addr==0x3F1C) addr = 0x3F0C;
	if (addr>0x2FFF && addr<0x3F00) addr = 0x3000 | (addr & 0xEFF);
	if (addr>=0x2000 && addr<0x2400) addr = (addr - 0x2000) + PPU->ntmap[0];
		else if (addr>=0x2400 && addr<0x2800) addr = (addr - 0x2400) + PPU->ntmap[1];
		else if (addr>=0x2800 && addr<0x2C00) addr = (addr - 0x2800) + PPU->ntmap[2];
		else if (addr>=0x2C00 && addr<0x3000) addr = (addr - 0x2C00) + PPU->ntmap[3];
	lastaddr = addr;
   if (addr<0x2000) {
		tempbyte = cartridge.CHRbank[addr>>10][addr&1023];
		if (cartridge.mapper == 9) {
			switch (addr & 0x3FF0) {
				case 0x0FD0: map9->latch1 = 0xFD; CHRswap(&cartridge, 0, map9->latch0_fd, 4096); break;
				case 0x0FE0: map9->latch1 = 0xFE; CHRswap(&cartridge, 0, map9->latch0_fe, 4096); break;
				case 0x1FD0: map9->latch2 = 0xFD; CHRswap(&cartridge, 1, map9->latch1_fd, 4096); break;
				case 0x1FE0: map9->latch2 = 0xFE; CHRswap(&cartridge, 1, map9->latch1_fe, 4096); break;
			}
		}
		oldbyte = PPU->bytebuffer;
		PPU->bytebuffer = tempbyte;
		return(oldbyte);
   }
   if (addr >= 0x3F00) tempbyte = VRAM[addr];
      else tempbyte = PPU->bytebuffer;
   PPU->bytebuffer = VRAM[addr];
   return(tempbyte);
} 

void writePPU(uint16_t addr, uint8_t value){
	addr &= 0x3FFF;
	//if (!(lastaddr & 0x1000) && (addr & 0x1000)) map4irqdecrement();

	if (addr>=0x3F00) addr = 0x3F00 | (addr & 0x1F);
	if (addr==0x3F10) addr = 0x3F00;
	if (addr==0x3F14) addr = 0x3F04;
	if (addr==0x3F18) addr = 0x3F08;
	if (addr==0x3F1C) addr = 0x3F0C;
	if (addr>0x2FFF && addr<0x3F00) addr = 0x3000 | (addr & 0xEFF);
	if (addr>=0x2000 && addr<0x2400) addr = (addr - 0x2000) + PPU->ntmap[0];
		else if (addr>=0x2400 && addr<0x2800) addr = (addr - 0x2400) + PPU->ntmap[1];
		else if (addr>=0x2800 && addr<0x2C00) addr = (addr - 0x2800) + PPU->ntmap[2];
		else if (addr>=0x2C00 && addr<0x3000) addr = (addr - 0x2C00) + PPU->ntmap[3];

	if (addr>=0x2000) VRAM[addr] = value;
		else if (cartridge.CHRcount==0) cartridge.CHRbank[addr>>10][addr&1023] = value;
	lastaddr = addr;
}

uint8_t lastwritten;
uint8_t readPPUregs(uint16_t addr) {
     uint8_t tempbyte;
     switch (addr) {
       case 0x2002:
		    tempbyte = (PPU->vblank << 7) | (PPU->sprzero << 6) | (PPU->sprover << 5) | (lastwritten & 0x1F);
            PPU->vblank = 0;
            PPU->addrlatch = 0;
            PPU->scrolllatch = 0;
            return(tempbyte);
       case 0x2004:
            return(OAM->RAM[OAM->addr]);
       case 0x2007:
			tempbyte = readPPU(PPU->addr);
            PPU->addr = (PPU->addr + PPU->addrinc) & 0x3FFF;
            return(tempbyte);
     }
     return(0);
}

void writePPUregs(uint16_t addr, uint8_t value) {
     PPU->regs[addr & 7] = value;
	 lastwritten = value;
     switch (addr) {
       case 0x2000:
            if (value&128) PPU->nmivblank = 1; else PPU->nmivblank = 0;
            if (value&32) PPU->sprsize = 16; else PPU->sprsize = 8;
            if (value&16) PPU->bgtable = 0x1000; else PPU->bgtable = 0x0000;
            if (value&8) PPU->sprtable = 0x1000; else PPU->sprtable = 0x0000;
            if (value&4) PPU->addrinc = 32; else PPU->addrinc = 1;
            PPU->nametable = value&3;
			PPU->tempaddr = (PPU->tempaddr & 0xF3FF) | (((uint16_t)value & 3) << 10);
            break;
       case 0x2001:
            if (value&16) PPU->sprvisible = 1; else PPU->sprvisible = 0;
            if (value&8) PPU->bgvisible = 1; else PPU->bgvisible = 0;
            if (value&4) PPU->sprclip = 0; else PPU->sprclip = 1;
            if (value&2) PPU->bgclip = 0; else PPU->bgclip = 1;
            break;
       case 0x2003:
            OAM->addr = value;
            break;
       case 0x2004:
            OAM->RAM[OAM->addr++] = value;
            break;
       case 0x2005:
            if (PPU->addrlatch == 0) {
              PPU->xscroll = value & 7;
			  PPU->tempaddr = (PPU->tempaddr & 0xFFE0) | ((uint16_t)value >> 3);
              PPU->addrlatch = 1;
            } else {
              PPU->yscroll = value & 7;
			  PPU->tempaddr = (PPU->tempaddr & 0xFC1F) | (((uint16_t)value & 0xF8) << 2);
              PPU->addrlatch = 0;
            }
			break;
       case 0x2006:
            if (PPU->addrlatch == 0) {
				PPU->tempaddr = (PPU->tempaddr & 0x00FF) | (((uint16_t)value & 0x3F) << 8);
				PPU->addrlatch = 1;
            } else {
				PPU->tempaddr = (PPU->tempaddr & 0xFF00) | (uint16_t)value;
				PPU->addr = PPU->tempaddr;
				PPU->addrlatch = 0;
            }
			break;
       case 0x2007:
            writePPU(PPU->addr, value);
            PPU->addr = (PPU->addr + PPU->addrinc) & 0x3FFF;
            break;
     }
}

uint16_t backgnd[256], sprfront[256], sprback[256], sprfrontidx[256], sprbackidx[256], sprzero[256];
extern uint16_t sprtablesave;

void rendersprites(uint16_t scanline) {
	uint16_t OAMptr, evalidx, attr, spry, sprx, table, tile, flipx, flipy, x, startx, plotx, calcx, calcy, patoffset;
	uint8_t curpixel, palette, priority;
	uint8_t valid[8] = { 0, 0, 0, 0, 0, 0, 0, 0 };
	uint8_t idxval[8];
	uint8_t spr0idx = 255;

	if (scanline == 0) return;
		else scanline--;

	OAMptr = 0;
	for (evalidx=0; evalidx<256; evalidx+=4) {
		if ((scanline >= OAM->RAM[evalidx]) && (scanline < (OAM->RAM[evalidx]  + PPU->sprsize))) {
			memcpy(&OAM->buf[OAMptr], &OAM->RAM[evalidx], 4);
			valid[OAMptr>>2] = 1;
			idxval[OAMptr>>2] = evalidx>>2;
			if (evalidx == 0) spr0idx = (uint8_t)OAMptr;
			OAMptr += 4;
		}
		if (OAMptr == 32) {
			PPU->sprover = 1;
			break;
		}
			if (cartridge.mapper == 9) {
				switch (OAM->RAM[evalidx+1]) {
					case 0xFD:
						if (PPU->sprtable == 0x0000) {
							CHRswap(&cartridge, 0, map9->latch0_fd, 4096);
						} else {
							CHRswap(&cartridge, 1, map9->latch1_fd, 4096);
						}
						break;
					case 0xFE:
						if (PPU->sprtable == 0x0000) {
							CHRswap(&cartridge, 0, map9->latch0_fe, 4096);
						} else {
							CHRswap(&cartridge, 1, map9->latch1_fe, 4096);
						}
						break;
				}
			}
	}

    if (PPU->sprclip) startx = 8;
		else startx = 0;
	if (PPU->sprsize == 8) table = PPU->sprtable;

	for (OAMptr=28; OAMptr<32; OAMptr-=4) {
		if (valid[OAMptr>>2] == 1) {
			spry = OAM->buf[OAMptr];
			spry = scanline - spry;
			tile = OAM->buf[OAMptr+1];
			attr = OAM->buf[OAMptr+2];
			sprx = OAM->buf[OAMptr+3];
			palette = (attr & 3) << 2;
			priority = (attr >> 5) & 1;
			flipx = (attr >> 6) & 1;
			flipy = (attr >> 7) & 1;

			if (PPU->sprsize == 16) {
				table = (tile & 1) << 12;
				tile &= 0xFE;
				if (flipy) calcy = (~spry & 15);
					else calcy = spry & 15;
				if (calcy > 7) tile++;
				calcy &= 7;
			} else {
				if (flipy) calcy = (~spry & 7);
					else calcy = spry & 7;
			}
			
			for (x=0; x<8; x++) {
				if (flipx) calcx = ~x & 7;
					else calcx = x;
				plotx = sprx + x;
				if ((plotx >= startx) && (plotx < 255)) {
					patoffset = table + (tile << 4) + calcy;
					curpixel = (cartridge.CHRbank[patoffset>>10][patoffset&1023] >> (~calcx & 7)) & 1; patoffset += 8;
					curpixel |= ((cartridge.CHRbank[patoffset>>10][patoffset&1023] >> (~calcx & 7)) & 1) << 1;
					if (priority) sprbackidx[plotx] = OAMptr >> 2;
						else sprfrontidx[plotx] = OAMptr >> 2;
					if (curpixel > 0) {
						if (spr0idx == OAMptr) if (backgnd[plotx] > 0) PPU->sprzero = 1;
						curpixel |= palette;
						if (priority) sprback[plotx] = 0x3F10 + (uint16_t)curpixel;
							else sprfront[plotx] = 0x3F10 + (uint16_t)curpixel;
					}
				}
			}

			if (cartridge.mapper == 9) {
				if (tile == 0xFD) {
					if (table) CHRswap(&cartridge, 1, map9->latch1_fd, 4096);
						else CHRswap(&cartridge, 0, map9->latch0_fd, 4096);
				} else if (tile == 0xFE) {
					if (table) CHRswap(&cartridge, 1, map9->latch1_fe, 4096);
						else CHRswap(&cartridge, 0, map9->latch0_fe, 4096);
				}
			}

		}
	}
}

extern uint64_t exec6502(uint64_t tickcount);
void renderbackground(uint16_t scanline) {
	uint16_t x, startx, calcx, calcy, patoffset, usent, ntbase, tile, lastx, lasty, tempval, patbank;
	uint8_t curpixel, curattrib;

	lasty = lastx = 0xFFFF;

	if (PPU->bgclip) startx = 8; //if background clipping enabled, don't draw leftmost 8 pixels
		else startx = 0; //otherwise do

		for (x=0; x<264; x++) {
			//if (!(x%3)) exec6502(1);

			calcx = ((PPU->addr << 3) & 0xFF) | PPU->xscroll;
			calcy = ((PPU->addr >> 2) & 0xF8) | PPU->yscroll;
			usent = (PPU->addr >> 10) & 3;

			ntbase = PPU->ntmap[usent]; //nametable base address
			tile = VRAM[ntbase + ((calcy & 248) << 2) + (calcx >> 3)]; //calculate tile offset based on X,Y coords
			patoffset = PPU->bgtable + (tile << 4) + (calcy & 7); //then turn that into the byte offset in the nametable array
			curattrib = (VRAM[ntbase + 0x3C0 + ((calcy & 224) >> 2) + (calcx >> 5)] >> (((calcx & 16) >> 3) | ((calcy & 16) >> 2)) & 3) << 2;
			patbank = patoffset >> 10;

			curpixel = (cartridge.CHRbank[patoffset>>10][patoffset&1023] >> (~calcx & 7)) & 1; //used to go fetch the two
			curpixel |= ((cartridge.CHRbank[patoffset>>10][(patoffset+8)&1023] >> (~calcx & 7)) & 1) << 1; //lower bits of the pixel's palette index

			if ((curpixel > 0) && (x >= startx) && (x < 256)) { //if those are not zero, then this is a visible pixel
				backgnd[x] = 0x3F00 + (curpixel | curattrib); //stick it in the buffer to be drawn
			}

			if (cartridge.mapper == 9) {
				if (tile == 0xFD) {
					if (patbank > 3) CHRswap(&cartridge, 1, map9->latch1_fd, 4096);
						else CHRswap(&cartridge, 0, map9->latch0_fd, 4096);
				} else if (tile == 0xFE) {
					if (patbank > 3) CHRswap(&cartridge, 1, map9->latch1_fe, 4096);
						else CHRswap(&cartridge, 0, map9->latch0_fe, 4096);
				}
			}

			PPU->xscroll++;
			if (PPU->xscroll == 8) {
				PPU->xscroll = 0;
				tempval = (PPU->addr & 0x1F) + 1;
				if (tempval == 32) PPU->addr ^= 1024;
				PPU->addr = (PPU->addr & 0xFFE0) | (tempval & 0x1F);
			}
		}
		PPU->yscroll++;
		if (PPU->yscroll == 8) {
			PPU->yscroll = 0;
			tempval = ((PPU->addr >> 5) & 0x1F) + 1;
			if (tempval == 30) {
				tempval = 0;
				PPU->addr ^= 2048;
			}
			PPU->addr = (PPU->addr & 0xFC1F) | ((tempval & 0x1F) << 5);
		}
}

void renderscanline(uint16_t scanline) {
     uint16_t tmpx;

	 if (scanline == 0) return;
	 scanline--;

	 PPU->sprover = 0;

	 memset(&backgnd[0], 0, sizeof(backgnd));
	 memset(&sprback[0], 0, sizeof(sprback));
	 memset(&sprbackidx[0], 0, sizeof(sprbackidx));
	 memset(&sprfront[0], 0, sizeof(sprfront));
	 memset(&sprfrontidx[0], 0, sizeof(sprfrontidx));
	 memset(&sprzero[0], 0, sizeof(sprzero));
	 memset(&OAM->buf[0], 0xFF, sizeof(OAM->buf));

	 if (PPU->bgvisible) renderbackground(scanline);
	 if (PPU->sprvisible) rendersprites(scanline);

	 for (tmpx=0; tmpx<256; tmpx++) {
       if (sprback[tmpx]==0) outputNES[scanline][tmpx] = VRAM[0x3F00] & 0x3F;
         else outputNES[scanline][tmpx] = VRAM[sprback[tmpx]] & 0x3F;
	   if (backgnd[tmpx]>0) {
			 outputNES[scanline][tmpx] = VRAM[backgnd[tmpx]] & 0x3F;
			 if (sprzero[tmpx]) PPU->sprzero = 1;
	   }
	   if (sprfront[tmpx]>0) outputNES[scanline][tmpx] = VRAM[sprfront[tmpx]] & 0x3F;
	   //if ((sprfront[tmpx]>0) && (sprfrontidx[tmpx] > sprbackidx[tmpx]))
		   //outputNES[scanline][tmpx] = VRAM[sprfront[tmpx]] & 0x3F;
	 }
}

void initPPU() {
	PPU = malloc(sizeof(struct PPU_s));
	VRAM = malloc(16384);
	if ((PPU == NULL) || (VRAM == NULL)) fatalerr("Unable to allocate memory for PPU!");
	OAM = malloc(sizeof(struct OAM_s));
	if (OAM == NULL) fatalerr("Unable to allocate memory for OAM!");

	memset(PPU, 0, sizeof(*PPU));
	memset(OAM, 0, sizeof(*OAM));
	memset(VRAM, 0, 16384);
	PPU->sprsize = 8;
}

void killPPU() {
	if (PPU != NULL) free(PPU);
	if (VRAM != NULL) free(VRAM);
	if (OAM != NULL) free(OAM);
}
