/* MoarNES source - avirecord.c

   This open-source software is (c)2010-2012 Mike Chambers, and is released
   under the terms of the GNU GPL v2 license.

   This file contains the code to record the emulator's audio/video output
   directly to an XViD AVI file. This makes use of the Revel library and the
   XViD core library. Both of these are statically linked into MoarNES.
*/

#include "config.h"
#include <stdio.h>
#include "revel.h"
#include "xvid.h"
#include "common.h"
#include "stdint.h"

extern void msgbox(uint8_t *msgtxt, uint16_t buttons);

extern SDL_AudioSpec audio;
extern uint16_t samplesperframe;
extern uint16_t avibufptr;
extern const uint32_t paletteNES[66];
uint32_t avibuffer[240][256];
uint8_t dorecord = 0;

Revel_Params revParams;
Revel_VideoFrame revFrame;
int16_t aviHandle;

void avistop();

void aviencodeframe(uint8_t *NESbuffer, uint8_t *audiobuffer, uint16_t samplecount) {
	Revel_Error revErr;
	uint16_t x, y, linear = 0;
	uint8_t stretchbuf[8192];

	for (y=0; y<240; y++) { //convert 8-bit palettized NES frame
		for (x=0; x<256; x++) { //to 32-bit BGRA array for XViD encoding
			avibuffer[y][x] = 0xFF000000 | paletteNES[NESbuffer[linear++]];
		}
	}

	revErr = Revel_EncodeFrame(aviHandle, &revFrame, NULL);
	if (revErr != REVEL_ERR_NONE) {
		msgbox("Error encoding AVI frame!", MB_ICONINFORMATION | MB_OK);
		avistop();
		return;
	}

	for (linear=0; linear<samplesperframe; linear++) {
		stretchbuf[linear] = audiobuffer[ (uint16_t)(((double)avibufptr / (double)samplesperframe) * (double)linear) ];
	}

	revErr = Revel_EncodeAudio(aviHandle, &stretchbuf[0], samplecount, NULL);
	if (revErr != REVEL_ERR_NONE) {
		msgbox("Error encoding AVI frame!", MB_ICONINFORMATION | MB_OK);
		avistop();
		return;
	}
}

void avistart(uint8_t *filename) {
	Revel_Error revErr;

	if (dorecord) return;

	revErr = Revel_CreateEncoder((int *)&aviHandle);
	if (revErr != REVEL_ERR_NONE) {
		msgbox("Error attempting to create an AVI encoder!", MB_ICONINFORMATION | MB_OK);
		return;
	}

	Revel_InitializeParams(&revParams);

	revParams.audioBits = 8;
	revParams.audioChannels = 1;
	revParams.audioRate = audio.freq;
	revParams.audioSampleFormat = REVEL_ASF_PCM;
	revParams.hasAudio = 1;
	
	revParams.codec = REVEL_CD_XVID;
	revParams.frameRate = 60.0f;
	revParams.height = 240;
	revParams.width = 256;
	revParams.quality = 1.0f;

	revFrame.bytesPerPixel = 4;
	revFrame.pixelFormat = REVEL_PF_BGRA;
	revFrame.pixels = &avibuffer[0][0];
	revFrame.width = 256;
	revFrame.height = 240;

	revErr = Revel_EncodeStart(aviHandle, filename, &revParams);
	if (revErr != REVEL_ERR_NONE) {
		msgbox("Error attempting to initalize AVI recording!", MB_ICONINFORMATION | MB_OK);
		return;
	} else dorecord = 1;
}

void avistop() {
	Revel_Error revErr;

	revErr = Revel_EncodeEnd(aviHandle, NULL);
	if (revErr != REVEL_ERR_NONE) {
		msgbox("Error attempting to stop AVI recording!", MB_ICONINFORMATION | MB_OK);
		return;
	} else dorecord = 0;
}
