/* MoarNES source - input.c

   This open-source software is (c)2010-2012 Mike Chambers, and is released
   under the terms of the GNU GPL v2 license.

   This file contains the code to configure and handle input for the NES gamepad
   emulation. They can be either keyboard or a joystick/gamepad.
*/

#include <stdio.h>
#include "common.h"
#include "stdint.h"
#include "input.h"

struct inputdev_s gamepad[2];
struct joystick_s *joy[32];
uint8_t joycount = 0;
uint8_t netsynced = 1;

extern void nmi6502();
extern void irq6502();
extern void reset6502();
extern void verifyexit();
extern void setscreenmode(uint16_t modeval);
extern uint8_t showdump;
extern bool running;
extern bool paused;
extern bool exitapp;
extern uint16_t screenmode;
extern uint16_t lastscreenmode;
extern uint8_t outputNES[240][256];
extern void blitframe(uint8_t *NESbuffer);
extern void savestate();
extern void loadstate();
extern bool dosavestate;
extern bool doloadstate;
extern uint8_t savestatefilename[MAX_PATH];
extern uint8_t app_path[MAX_PATH];
extern bool isquicksave;
extern uint32_t saveslot;

uint8_t joyread(uint8_t joynum) {
	uint8_t buttondata;
	buttondata = gamepad[joynum].gamepad.buttonvalue[gamepad[joynum].currentread];
	if (buttondata != 1) buttondata = 0;
	gamepad[joynum].currentread = (gamepad[joynum].currentread + 1) & 7;
	return(buttondata);
}

void joystrobe() {
	gamepad[0].currentread = 0;
	gamepad[1].currentread = 0;
}

void pressup(uint8_t controller, uint8_t eventtype) {
	if (!netsynced) return;
	if (eventtype == INPUT_EVENT_PRESS) {
		gamepad[controller].gamepad.buttonname.up = 1;
		if (gamepad[controller].gamepad.buttonname.down == 1) gamepad[controller].gamepad.buttonname.down = 2;
	} else if (eventtype == INPUT_EVENT_UNPRESS) {
		gamepad[controller].gamepad.buttonname.up = 0;
		if (gamepad[controller].gamepad.buttonname.down == 2) gamepad[controller].gamepad.buttonname.down = 1;
	}
}

void pressdown(uint8_t controller, uint8_t eventtype) {
	if (!netsynced) return;
	if (eventtype == INPUT_EVENT_PRESS) {
		gamepad[controller].gamepad.buttonname.down = 1;
		if (gamepad[controller].gamepad.buttonname.up == 1) gamepad[controller].gamepad.buttonname.up = 2;
	} else if (eventtype == INPUT_EVENT_UNPRESS) {
		gamepad[controller].gamepad.buttonname.down = 0;
		if (gamepad[controller].gamepad.buttonname.up == 2) gamepad[controller].gamepad.buttonname.up = 1;
	}
}

void pressleft(uint8_t controller, uint8_t eventtype) {
	if (!netsynced) return;
	if (eventtype == INPUT_EVENT_PRESS) {
		gamepad[controller].gamepad.buttonname.left = 1;
		if (gamepad[controller].gamepad.buttonname.right == 1) gamepad[controller].gamepad.buttonname.right = 2;
	} else if (eventtype == INPUT_EVENT_UNPRESS) {
		gamepad[controller].gamepad.buttonname.left = 0;
		if (gamepad[controller].gamepad.buttonname.right == 2) gamepad[controller].gamepad.buttonname.right = 1;
	}
}

void pressright(uint8_t controller, uint8_t eventtype) {
	if (!netsynced) return;
	if (eventtype == INPUT_EVENT_PRESS) {
		gamepad[controller].gamepad.buttonname.right = 1;
		if (gamepad[controller].gamepad.buttonname.left == 1) gamepad[controller].gamepad.buttonname.left = 2;
	} else if (eventtype == INPUT_EVENT_UNPRESS) {
		gamepad[controller].gamepad.buttonname.right = 0;
		if (gamepad[controller].gamepad.buttonname.left == 2) gamepad[controller].gamepad.buttonname.left = 1;
	}
}

void pressA(uint8_t controller, uint8_t eventtype) {
	if (!netsynced) return;
	if (eventtype == INPUT_EVENT_PRESS) {
		gamepad[controller].gamepad.buttonname.A = 1;
	} else if (eventtype == INPUT_EVENT_UNPRESS) {
		gamepad[controller].gamepad.buttonname.A = 0;
	}
}

void pressB(uint8_t controller, uint8_t eventtype) {
	if (!netsynced) return;
	if (eventtype == INPUT_EVENT_PRESS) {
		gamepad[controller].gamepad.buttonname.B = 1;
	} else if (eventtype == INPUT_EVENT_UNPRESS) {
		gamepad[controller].gamepad.buttonname.B = 0;
	}
}

void pressstart(uint8_t controller, uint8_t eventtype) {
	if (!netsynced) return;
	if (eventtype == INPUT_EVENT_PRESS) {
		gamepad[controller].gamepad.buttonname.start = 1;
	} else if (eventtype == INPUT_EVENT_UNPRESS) {
		gamepad[controller].gamepad.buttonname.start = 0;
	}
}

void pressselect(uint8_t controller, uint8_t eventtype) {
	if (!netsynced) return;
	if (eventtype == INPUT_EVENT_PRESS) {
		gamepad[controller].gamepad.buttonname.select = 1;
	} else if (eventtype == INPUT_EVENT_UNPRESS) {
		gamepad[controller].gamepad.buttonname.select = 0;
	}
}

void pressjoy(uint8_t controller, uint8_t pressval) {
	if (!netsynced) return;
	switch (pressval) {
		case 0: pressA(controller, INPUT_EVENT_PRESS); break;
		case 1: pressB(controller, INPUT_EVENT_PRESS); break;
		case 2: pressselect(controller, INPUT_EVENT_PRESS); break;
		case 3: pressstart(controller, INPUT_EVENT_PRESS); break;
		case 4: pressup(controller, INPUT_EVENT_PRESS); break;
		case 5: pressdown(controller, INPUT_EVENT_PRESS); break;
		case 6: pressleft(controller, INPUT_EVENT_PRESS); break;
		case 7: pressright(controller, INPUT_EVENT_PRESS); break;
	}
}

void unpressjoy(uint8_t controller, uint8_t pressval) {
	if (!netsynced) return;
	switch (pressval) {
		case 0: pressA(controller, INPUT_EVENT_UNPRESS); break;
		case 1: pressB(controller, INPUT_EVENT_UNPRESS); break;
		case 2: pressselect(controller, INPUT_EVENT_UNPRESS); break;
		case 3: pressstart(controller, INPUT_EVENT_UNPRESS); break;
		case 4: pressup(controller, INPUT_EVENT_UNPRESS); break;
		case 5: pressdown(controller, INPUT_EVENT_UNPRESS); break;
		case 6: pressleft(controller, INPUT_EVENT_UNPRESS); break;
		case 7: pressright(controller, INPUT_EVENT_UNPRESS); break;
	}
}

void processinput() {
	SDL_Event event;
	uint8_t inputcheck, curcontroller;

	while (SDL_PollEvent(&event)) {
		switch (event.type) {
			case SDL_KEYDOWN:
				for (curcontroller=0; curcontroller<2; curcontroller++) if (gamepad[curcontroller].inputmethod == INPUT_METHOD_KEYBOARD) {
					for (inputcheck=0; inputcheck<8; inputcheck++) {
						if (event.key.keysym.sym == gamepad[curcontroller].mapping[inputcheck]) {
							switch (inputcheck) {
								case 0: pressA(curcontroller, INPUT_EVENT_PRESS); break; //A
								case 1: pressB(curcontroller, INPUT_EVENT_PRESS); break; //B
								case 2: pressselect(curcontroller, INPUT_EVENT_PRESS); break; //select
								case 3: pressstart(curcontroller, INPUT_EVENT_PRESS); break; //start
								case 4: pressup(curcontroller, INPUT_EVENT_PRESS); break; //up
								case 5: pressdown(curcontroller, INPUT_EVENT_PRESS); break; //down
								case 6: pressleft(curcontroller, INPUT_EVENT_PRESS); break; //left
								case 7: pressright(curcontroller, INPUT_EVENT_PRESS); break; //right
							}
						}
					}
				}
				switch (event.key.keysym.sym) {
					case SDLK_F1:
						sprintf(&savestatefilename[0], "%sslot1.mns", &app_path[0]);
						dosavestate = true;
						isquicksave = true;
						saveslot = 0;
						break;
					case SDLK_F2:
						sprintf(&savestatefilename[0], "%sslot2.mns", &app_path[0]);
						dosavestate = true;
						isquicksave = true;
						saveslot = 1;
						break;
					case SDLK_F3:
						sprintf(&savestatefilename[0], "%sslot3.mns", &app_path[0]);
						dosavestate = true;
						isquicksave = true;
						saveslot = 2;
						break;
					case SDLK_F4:
						sprintf(&savestatefilename[0], "%sslot4.mns", &app_path[0]);
						dosavestate = true;
						isquicksave = true;
						saveslot = 3;
						break;
					case SDLK_F5:
						sprintf(&savestatefilename[0], "%sslot1.mns", &app_path[0]);
						doloadstate = true;
						isquicksave = true;
						saveslot = 0;
						break;
					case SDLK_F6:
						sprintf(&savestatefilename[0], "%sslot2.mns", &app_path[0]);
						doloadstate = true;
						isquicksave = true;
						saveslot = 1;
						break;
					case SDLK_F7:
						sprintf(&savestatefilename[0], "%sslot3.mns", &app_path[0]);
						doloadstate = true;
						isquicksave = true;
						saveslot = 2;
						break;
					case SDLK_F8:
						sprintf(&savestatefilename[0], "%sslot4.mns", &app_path[0]);
						doloadstate = true;
						isquicksave = true;
						saveslot = 3;
						break;
					case SDLK_ESCAPE:
						if (screenmode == 5) setscreenmode(lastscreenmode);
						break;
				}
				break;

			case SDL_KEYUP:
				for (curcontroller=0; curcontroller<2; curcontroller++) if (gamepad[curcontroller].inputmethod == INPUT_METHOD_KEYBOARD) {
					for (inputcheck=0; inputcheck<8; inputcheck++) {
						if (event.key.keysym.sym == gamepad[curcontroller].mapping[inputcheck]) {
							switch (inputcheck) {
								case 0: pressA(curcontroller, INPUT_EVENT_UNPRESS); break; //A
								case 1: pressB(curcontroller, INPUT_EVENT_UNPRESS); break; //B
								case 2: pressselect(curcontroller, INPUT_EVENT_UNPRESS); break; //select
								case 3: pressstart(curcontroller, INPUT_EVENT_UNPRESS); break; //start
								case 4: pressup(curcontroller, INPUT_EVENT_UNPRESS); break; //up
								case 5: pressdown(curcontroller, INPUT_EVENT_UNPRESS); break; //down
								case 6: pressleft(curcontroller, INPUT_EVENT_UNPRESS); break; //left
								case 7: pressright(curcontroller, INPUT_EVENT_UNPRESS); break; //right
							}
						}
					}
				}
				break;

			case SDL_JOYHATMOTION:
				for (curcontroller=0; curcontroller<2; curcontroller++) if (gamepad[curcontroller].inputmethod == INPUT_METHOD_JOYSTICK) {
					for (inputcheck=0; inputcheck<8; inputcheck++) {
						if (event.jhat.which == gamepad[curcontroller].joynum[inputcheck]) {
							if (event.jhat.hat == gamepad[curcontroller].hatmap[inputcheck]) {
								if (event.jhat.value & gamepad[curcontroller].mapping[inputcheck]) {
									pressjoy(curcontroller, inputcheck);
								} else {
									unpressjoy(curcontroller, inputcheck);
								}
							}
						}
					}
				}
				break;

			case SDL_JOYBUTTONDOWN:
				for (curcontroller=0; curcontroller<2; curcontroller++) if (gamepad[curcontroller].inputmethod == INPUT_METHOD_JOYSTICK) {
					for (inputcheck=0; inputcheck<8; inputcheck++) {
						if (event.jbutton.which == gamepad[curcontroller].joynum[inputcheck]) {
							if (gamepad[curcontroller].hatmap[inputcheck] == 0xFF) {
								if (event.jbutton.button == gamepad[curcontroller].mapping[inputcheck]) {
									pressjoy(curcontroller, inputcheck);
								}
							}
						}
					}
				}
				break;

			case SDL_JOYBUTTONUP:
				for (curcontroller=0; curcontroller<2; curcontroller++) if (gamepad[curcontroller].inputmethod == INPUT_METHOD_JOYSTICK) {
					for (inputcheck=0; inputcheck<8; inputcheck++) {
						if (event.jbutton.which == gamepad[curcontroller].joynum[inputcheck]) {
							if (gamepad[curcontroller].hatmap[inputcheck] == 0xFF) {
								if (event.jbutton.button == gamepad[curcontroller].mapping[inputcheck]) {
									unpressjoy(curcontroller, inputcheck);
								}
							}
						}
					}
				}
				break;

			case SDL_QUIT:
				verifyexit();
				break;
		}
	}
}

uint16_t waitforkey() {
	SDL_Event event;

	for (;;) {
		while (SDL_PollEvent(&event)) {
			switch (event.type) {
				case SDL_KEYDOWN:
					return(event.key.keysym.sym);
					break;
				case SDL_QUIT:
					verifyexit();
					break;
			}
		}
		SDL_Delay(10); //free up CPU time
	}
}

uint32_t waitforjoy() {
	SDL_Event event;
	uint32_t result;

	for (;;) {
		while (SDL_PollEvent(&event)) {
			switch (event.type) {
				case SDL_JOYHATMOTION:
					if (event.jhat.value != SDL_HAT_CENTERED) {
						result = 0x80000000 | ((uint32_t)event.jhat.which << 16) | ((uint32_t)event.jhat.hat << 8) | (uint32_t)event.jhat.value;
						return(result);
					}
					break;
				case SDL_JOYBUTTONDOWN:
					if (event.jbutton.state == SDL_PRESSED) {
						result = ((uint32_t)event.jbutton.which << 16) | (uint32_t)event.jbutton.button;
						return(result);
					}
					break;
				case SDL_QUIT:
					verifyexit();
					break;
			}
		}
		SDL_Delay(10); //free up CPU time
	}
}

extern void drawtext(uint8_t *text, uint16_t x, uint16_t y);
void configurekeyboard(uint8_t controller) {
	SDL_Event event;
	uint16_t ypos = 0;
	uint8_t textdata[40];
	uint8_t screensave[240][256];

	memcpy(&screensave[0][0], &outputNES[0][0], 61440);
	memset(&outputNES[0][0], 65, 61440);

	while (SDL_PollEvent(&event)) { } //flush any waiting keystrokes

	gamepad[controller].inputmethod = INPUT_METHOD_KEYBOARD;
	sprintf(&textdata[0], "Gamepad %u key configuration", controller + 1);
	drawtext(&textdata[0], 0, ypos); ypos += 16;
	drawtext("Press new key for A", 0, ypos); ypos += 8; blitframe(&outputNES[0][0]);
	gamepad[controller].mapping[0] = waitforkey();
	drawtext("Press new key for B", 0, ypos); ypos += 8; blitframe(&outputNES[0][0]);
	gamepad[controller].mapping[1] = waitforkey();
	drawtext("Press new key for select", 0, ypos); ypos += 8; blitframe(&outputNES[0][0]);
	gamepad[controller].mapping[2] = waitforkey();
	drawtext("Press new key for start", 0, ypos); ypos += 8; blitframe(&outputNES[0][0]);
	gamepad[controller].mapping[3] = waitforkey();
	drawtext("Press new key for up", 0, ypos); ypos += 8; blitframe(&outputNES[0][0]);
	gamepad[controller].mapping[4] = waitforkey();
	drawtext("Press new key for down", 0, ypos); ypos += 8; blitframe(&outputNES[0][0]);
	gamepad[controller].mapping[5] = waitforkey();
	drawtext("Press new key for left", 0, ypos); ypos += 8; blitframe(&outputNES[0][0]);
	gamepad[controller].mapping[6] = waitforkey();
	drawtext("Press new key for right", 0, ypos); ypos += 8; blitframe(&outputNES[0][0]);
	gamepad[controller].mapping[7] = waitforkey();

	memcpy(&outputNES[0][0], &screensave[0][0], 61440);
	blitframe(&outputNES[0][0]);
}

void configurejoystick(uint8_t controller) {
	SDL_Event event;
	uint16_t ypos = 0;
	uint8_t textdata[40], clearbutton;
	uint32_t result;
	uint8_t screensave[240][256];

	memcpy(&screensave[0][0], &outputNES[0][0], 61440);
	memset(&outputNES[0][0], 65, 61440);

	while (SDL_PollEvent(&event)) { } //flush any waiting keystrokes

	for (clearbutton=0; clearbutton<8; clearbutton++) {
		gamepad[controller].mapping[clearbutton] = 0xFFFF;
	}

	gamepad[controller].inputmethod = INPUT_METHOD_JOYSTICK;
	sprintf(&textdata[0], "Gamepad %u joystick configuration", controller + 1);
	drawtext(&textdata[0], 0, ypos); ypos += 16;
	drawtext("Press new key for A", 0, ypos); ypos += 8; blitframe(&outputNES[0][0]);
	result = waitforjoy(); if (result & 0x80000000) {
		gamepad[controller].hatmap[0] = (uint8_t)(result >> 8);
		gamepad[controller].joynum[0] = (uint8_t)(result >> 16);
		gamepad[controller].mapping[0] = (uint8_t)result;
	} else {
		gamepad[controller].hatmap[0] = 0xFF;
		gamepad[controller].joynum[0] = (uint8_t)(result >> 16);
		gamepad[controller].mapping[0] = (uint8_t)result;
	}

	drawtext("Press new key for B", 0, ypos); ypos += 8; blitframe(&outputNES[0][0]);
	result = waitforjoy(); if (result & 0x80000000) {
		gamepad[controller].hatmap[1] = (uint8_t)(result >> 8);
		gamepad[controller].joynum[1] = (uint8_t)(result >> 16);
		gamepad[controller].mapping[1] = (uint8_t)result;
	} else {
		gamepad[controller].hatmap[1] = 0xFF;
		gamepad[controller].joynum[1] = (uint8_t)(result >> 16);
		gamepad[controller].mapping[1] = (uint8_t)result;
	}

	drawtext("Press new key for select", 0, ypos); ypos += 8; blitframe(&outputNES[0][0]);
	result = waitforjoy(); if (result & 0x80000000) {
		gamepad[controller].hatmap[2] = (uint8_t)(result >> 8);
		gamepad[controller].joynum[2] = (uint8_t)(result >> 16);
		gamepad[controller].mapping[2] = (uint8_t)result;
	} else {
		gamepad[controller].hatmap[2] = 0xFF;
		gamepad[controller].joynum[2] = (uint8_t)(result >> 16);
		gamepad[controller].mapping[2] = (uint8_t)result;
	}

	drawtext("Press new key for start", 0, ypos); ypos += 8; blitframe(&outputNES[0][0]);
	result = waitforjoy(); if (result & 0x80000000) {
		gamepad[controller].hatmap[3] = (uint8_t)(result >> 8);
		gamepad[controller].joynum[3] = (uint8_t)(result >> 16);
		gamepad[controller].mapping[3] = (uint8_t)result;
	} else {
		gamepad[controller].hatmap[3] = 0xFF;
		gamepad[controller].joynum[3] = (uint8_t)(result >> 16);
		gamepad[controller].mapping[3] = (uint8_t)result;
	}

	drawtext("Press new key for up", 0, ypos); ypos += 8; blitframe(&outputNES[0][0]);
	result = waitforjoy(); if (result & 0x80000000) {
		gamepad[controller].hatmap[4] = (uint8_t)(result >> 8);
		gamepad[controller].joynum[4] = (uint8_t)(result >> 16);
		gamepad[controller].mapping[4] = (uint8_t)result;
	} else {
		gamepad[controller].hatmap[4] = 0xFF;
		gamepad[controller].joynum[4] = (uint8_t)(result >> 16);
		gamepad[controller].mapping[4] = (uint8_t)result;
	}

	drawtext("Press new key for down", 0, ypos); ypos += 8; blitframe(&outputNES[0][0]);
	result = waitforjoy(); if (result & 0x80000000) {
		gamepad[controller].hatmap[5] = (uint8_t)(result >> 8);
		gamepad[controller].joynum[5] = (uint8_t)(result >> 16);
		gamepad[controller].mapping[5] = (uint8_t)result;
	} else {
		gamepad[controller].hatmap[5] = 0xFF;
		gamepad[controller].joynum[5] = (uint8_t)(result >> 16);
		gamepad[controller].mapping[5] = (uint8_t)result;
	}

	drawtext("Press new key for left", 0, ypos); ypos += 8; blitframe(&outputNES[0][0]);
	result = waitforjoy(); if (result & 0x80000000) {
		gamepad[controller].hatmap[6] = (uint8_t)(result >> 8);
		gamepad[controller].joynum[6] = (uint8_t)(result >> 16);
		gamepad[controller].mapping[6] = (uint8_t)result;
	} else {
		gamepad[controller].hatmap[6] = 0xFF;
		gamepad[controller].joynum[6] = (uint8_t)(result >> 16);
		gamepad[controller].mapping[6] = (uint8_t)result;
	}

	drawtext("Press new key for right", 0, ypos); ypos += 8; blitframe(&outputNES[0][0]);
	result = waitforjoy(); if (result & 0x80000000) {
		gamepad[controller].hatmap[7] = (uint8_t)(result >> 8);
		gamepad[controller].joynum[7] = (uint8_t)(result >> 16);
		gamepad[controller].mapping[7] = (uint8_t)result;
	} else {
		gamepad[controller].hatmap[7] = 0xFF;
		gamepad[controller].joynum[7] = (uint8_t)(result >> 16);
		gamepad[controller].mapping[7] = (uint8_t)result;
	}

	memcpy(&outputNES[0][0], &screensave[0][0], 61440);
	blitframe(&outputNES[0][0]);
}

void initinput() {
	uint8_t initjoy;

	gamepad[0].inputmethod = INPUT_METHOD_KEYBOARD;
	gamepad[1].inputmethod = INPUT_METHOD_KEYBOARD;

	//default player 1 keyboard config
	gamepad[0].mapping[0] = SDLK_x; //A
	gamepad[0].mapping[1] = SDLK_z; //B
	gamepad[0].mapping[2] = SDLK_RSHIFT; //select
	gamepad[0].mapping[3] = SDLK_RETURN; //start
	gamepad[0].mapping[4] = SDLK_UP;
	gamepad[0].mapping[5] = SDLK_DOWN;
	gamepad[0].mapping[6] = SDLK_LEFT;
	gamepad[0].mapping[7] = SDLK_RIGHT;

	//initialize joystick stuff
	SDL_JoystickEventState(SDL_ENABLE);
	joycount = SDL_NumJoysticks();
	if (joycount < 0) { joycount = 0; return; }
	for (initjoy=0; (initjoy<joycount) && (joycount<32); initjoy++) {
		joy[initjoy] = malloc(sizeof(struct joystick_s));
		joy[initjoy]->sdljoy = SDL_JoystickOpen(initjoy);
		if (joy[initjoy]->sdljoy != NULL) {
			joy[initjoy]->numbuttons = SDL_JoystickNumButtons(joy[initjoy]->sdljoy);
			joy[initjoy]->numhats = SDL_JoystickNumHats(joy[initjoy]->sdljoy);
		}
	}
}
