/* MoarNES source - main.c

   This open-source software is (c)2010-2012 Mike Chambers, and is released
   under the terms of the GNU GPL v2 license.

   MoarNES is an emulator of the Nintendo Entertainment System (NES) game
   console that was very popular in the 1980's. It emulates the various
   components of the console, such as the CPU, graphics and audio processors,
   memory mapping of the system, etc.

   It makes use of the Simple Directmedia Layer library by Sam Latinga, for
   ease is porting this emulator to various platforms without needing to
   completely rewrite any low-level graphics, audio, or input code.

   All other code in this emulator is written completely by Mike Chambers.

   It can run dumped images of most of the original NES game cartridges. These are
   widely available on the internet.
*/


#include <stdio.h>
#include "config.h"
#include "common.h"
#include "stdint.h"
#include "input.h"

const uint8_t *build = "v0.12.11.2 alpha";
bool ROMwaiting = false;
bool ROMisNSF = false;
bool exitapp = false;
uint8_t nextROM[512];
uint8_t app_path[MAX_PATH];
uint8_t doaudio = 1;

#ifdef _WIN32
	HANDLE GameThread;
	HANDLE NESThread;
#endif
extern bool initSDL(uint8_t *windowcaption);
extern uint8_t *readSDLerr();
extern DWORD WINAPI threadNES(LPVOID lParam);
extern void execNES();
extern bool loadROM(uint8_t *filename);

extern void initPPU();
extern void killPPU();
extern void initinput();
extern void blitframe(uint8_t *NESbuffer);

extern void hookexternal(void *funcptr);
extern void processinput();
extern void reset6502();
extern bool initAPU(uint8_t CPUmode, uint32_t samplerate);
extern void tickchannelsAPU();
extern void drawtext(uint8_t *text, uint16_t, uint16_t y);

extern uint8_t dorecord;
extern uint8_t recordpath[256];
extern uint8_t outputNES[240][256];

extern bool doloadmovie;
extern bool dosavemovie;

extern bool running;
extern bool paused;
extern uint32_t cursamplerate;

extern int16_t msgbox(uint8_t *msgtext, uint16_t buttons);

extern uint8_t initnet();
#ifdef NETPLAY_ENABLE
	extern uint8_t setupudp(uint8_t mode, uint8_t *serverhost);
	extern void processnet(uint32_t cursync);
	extern uint32_t netplaysyncgap;
#endif

extern void avistop();

void loadconfig();
void saveconfig();

extern const uint32_t paletteNES[66];

int main(int argc, char *argv[]) {
	uint8_t textdata[256] = { 0 };
	int16_t charfind;
	FILE *pal = NULL;

	charfind = (int16_t)GetModuleFileName(NULL, &app_path[0], MAX_PATH);
	for (; charfind>=0; charfind--) {
		if (app_path[charfind] == '\\') {
			app_path[charfind+1] = 0;
			break;
		}
	}

	sprintf(&textdata[0], "MoarNES %s", build);
	if (initSDL(&textdata[0]) == false) return(-1);
	if (initAPU(1, 96000) == false) if (initAPU(1, 48000) == false) if (initAPU(1, 44100) == false)
		if (initAPU(1, 32000) == false) if (initAPU(1, 22050) == false) if (initAPU(1, 11025) == false) doaudio = 0;
	initPPU();

	hookexternal(&tickchannelsAPU);
	initinput();
	loadconfig();
	#ifdef NETPLAY_ENABLE
		if (initnet()) {
			/*if (argc > 2) {
				if (!memcmp(argv[1], "-server", 7)) setupudp(1, argv[2]);
					else if (!memcmp(argv[1], "-client", 7)) setupudp(2, argv[2]);
				if (argc > 3) netplaysyncgap = atoi(argv[3]);
			}*/
		}
	#endif

	if (argc > 1) {
		sprintf(&nextROM[0], argv[1]);
		ROMwaiting = true;
	}

	memset(&outputNES[0][0], 65, 61440); //blank screen to black
	drawtext(&textdata[0], (uint16_t)(128-(strlen(&textdata[0])*8)/2), 224);
	if (doaudio)
		sprintf(&textdata[0], "Audio initialized at %lu Hz", cursamplerate);
	else
		sprintf(&textdata[0], "Unable to initialize audio");
	drawtext(&textdata[0], (uint16_t)(128-(strlen(&textdata[0])*8)/2), 232);

	/*pal = fopen("C:\\Users\\Mike Chambers\\Documents\\My PSP Files\\Palettes\\nes.PspPalette", "w");
	fprintf(pal, "JASC-PAL\n0100\n64\n");
	for (xb=0; xb<64; xb++) {
		fprintf(pal, "%u %u %u\n", paletteNES[xb]>>16, (paletteNES[xb]>>8)&255, paletteNES[xb]&255);
		for (yb=0; yb<4; yb++) {
			outputNES[yb][(xb<<2)] = xb;
			outputNES[yb][(xb<<2)+1] = xb;
			outputNES[yb][(xb<<2)+2] = xb;
			outputNES[yb][(xb<<2)+3] = xb;
		}
	}
	fclose(pal);*/
	blitframe(&outputNES[0][0]);

	while(exitapp == false) {
		#ifdef NETPLAY_ENABLE
			processnet(0xAAAAAAAA);
		#endif
		processinput();
		if (running == true) {
			execNES();
			if (initAPU(1, 96000) == false) if (initAPU(1, 48000) == false) if (initAPU(1, 44100) == false)
				if (initAPU(1, 32000) == false) if (initAPU(1, 22050) == false) if (initAPU(1, 11025) == false) doaudio = 0;
		}
		if (ROMwaiting == true) {
			loadROM(&nextROM[0]);
			running = true;
		}
		if (running == false) processinput();
		SDL_Delay(10);
	}

	if (dorecord) avistop();

	saveconfig();
	return(0);
}

#ifdef _WIN32
extern HWND myWindow;
int16_t msgbox(uint8_t *msgtxt, uint16_t buttons) {
	return(MessageBox(myWindow, msgtxt, "MoarNES", buttons));
}

void fatalerr(uint8_t *errmsg) {
	uint8_t newmsg[1024];

	sprintf(&newmsg[0], "FATAL ERROR: %s", errmsg);
	MessageBox(myWindow, &newmsg[0], "Fatal error", MB_ICONERROR | MB_OK);
	exit(-1);
}
#endif

bool verifyexit() {
	switch (msgbox("Are you sure you want to exit MoarNES?", MB_ICONQUESTION | MB_YESNO)) {
		case IDYES:
			running = false;
			paused = false;
			exitapp = true;
			return(true);
	}

	return(false);
}

extern uint8_t app_path[MAX_PATH];
extern struct inputdev_s gamepad[2];
extern uint16_t screenmode;
extern void setscreenmode(uint16_t modeval);
#ifdef _WIN32
extern HWND myWindow;
RECT winrect;
#else
uint8_t blank[4] = { 0, 0, 0, 0 };
#endif
void loadconfig() {
	FILE *configfile = NULL;
	uint8_t path[MAX_PATH];
	int16_t wx, wy;

	sprintf(&path[0], "%sMoarNES.conf", &app_path[0]);
	configfile = fopen(&path[0], "rb");
	if (configfile == NULL) {
		msgbox("This appears to be the first time you have run MoarNES, or the configuration file has been deleted. A new one will be created in the application's folder using the default options.", MB_ICONINFORMATION | MB_OK);
		saveconfig();
		return;
	}

	fread(&screenmode, 1, 2, configfile);
	fread(&wx, 1, 2, configfile);
	fread(&wy, 1, 2, configfile);
	#ifdef _WIN32
		SetWindowPos(myWindow, NULL, wx, wy, 0, 0, SWP_NOSIZE);
	#endif
	fread(&gamepad[0], 1, sizeof(struct inputdev_s) << 1, configfile);
	fclose(configfile);

	setscreenmode(screenmode);
}

void saveconfig() {
	FILE *configfile = NULL;
	uint8_t path[MAX_PATH];

	sprintf(&path[0], "%sMoarNES.conf", &app_path[0]);
	configfile = fopen(&path[0], "wb");
	if (configfile == NULL) return;

	fwrite(&screenmode, 1, 2, configfile);
	#ifdef _WIN32
		GetWindowRect(myWindow, (LPRECT)&winrect);
		fwrite(&winrect.left, 1, 2, configfile);
		fwrite(&winrect.top, 1, 2, configfile);
	#else
		fwrite(&blank[0], 1, 4, configfile);
	#endif
	fwrite(&gamepad[0], 1, sizeof(struct inputdev_s) << 1, configfile);

	fclose(configfile);
}
