/* MoarNES source - netplay.c

   This open-source software is (c)2010-2012 Mike Chambers, and is released
   under the terms of the GNU GPL v2 license.

   This file contains the code to handle network play via UDP.
*/

#include "config.h"
#ifdef NETPLAY_ENABLE
#include <stdio.h>
#include "SDL_net.h"
#include "common.h"
#include "stdint.h"

extern int16_t msgbox(uint8_t *msgtxt, uint16_t buttons);

UDPsocket sock;
UDPpacket packet, outpacket;
uint8_t recvbuf[1024], sendbuf[1024];
IPaddress ipaddr, recvaddr;

uint8_t netmode = 0; //0 = netplay disabled, 1 = server, 2 = client

uint8_t initnet() {
	uint8_t errmsg[512];
	if (SDLNet_Init()==-1) {
		sprintf(&errmsg[0], "SDLNet_Init: %s\n", SDLNet_GetError());
		msgbox(&errmsg[0], MB_ICONINFORMATION| MB_OK);
		return(false);
	}

	return(true);
}

void stopnet() {
	SDLNet_Quit();
}

uint8_t setupudp(uint8_t mode, uint8_t *serverhost) {
    sock = SDLNet_UDP_Open(1337);

	if (!sock) {
		msgbox("UDP socket open failed!", MB_ICONINFORMATION | MB_OK);
		return(false);
	}

	if (mode == 2) {
		if (SDLNet_ResolveHost(&ipaddr, serverhost, 1337)) {
			msgbox("Failed to resolve host!", MB_ICONINFORMATION | MB_OK);
			return(false);
		}
	}

	netmode = mode;
	if (netmode == 1) SDLNet_ResolveHost(&ipaddr, serverhost, 1337);
	//	else if (netmode == 2) SDLNet_ResolveHost(&ipaddr, "192.168.1.91", 1337);

	return(true);
}

void processnet(uint32_t cursync) {
	if (netmode == 0) return;

	while (true) {
		packet.data = &recvbuf[0];
		packet.maxlen = 1024;
		packet.channel = -1;
		if (SDLNet_UDP_Recv(sock, &packet)) {
			//ipaddr = packet.address;
			if (!memcmp(&recvbuf[0], &cursync, 4)) break;
		} else break;
	}
}

void sendudp(uint8_t *data, int16_t len) {
	outpacket.address = ipaddr;
	outpacket.data = data;
	outpacket.len = len;
	outpacket.channel = -1;
	if (!SDLNet_UDP_Send(sock, -1, &outpacket)) { } //msgbox("Send error!", MB_OK);
}
#endif
